"""WireViz-Web - a wrapper around WireViz for bringing it to the web. Easily document cables and wiring harnesses."""
__appname__ = "wireviz-web"


# Single-sourcing the package version
# https://cjolowicz.github.io/posts/hypermodern-python-06-ci-cd/
try:
    from importlib.metadata import PackageNotFoundError, version  # noqa
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # noqa

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

import logging
import os

from flask import Flask
from flask_cors import CORS

logger = logging.getLogger(__name__)


def create_app(test_config=None) -> Flask:
    """
    Create and configure the Flask application, with CORS.

    - https://flask.palletsprojects.com/en/1.1.x/tutorial/factory/
    - https://flask-cors.readthedocs.io/

    :param test_config:
    :return: Configured Flask application.
    """

    logger.info("Init " + __appname__)
    # Create Flask application.
    app = Flask(__name__)

    # Initialize Cross Origin Resource sharing support for
    # the application on all routes, for all origins and methods.
    # enable cors if flag is set
    if os.getenv("CORS_ENABLE", "False").lower() in ("true", "1", "t"):
        CORS(app)
        # app.config['CORS_HEADERS'] = 'Content-Type'
        logger.info("CORS enabled for all routes")

    return app


"""
About
-----
`PlantUML Text Encoding format`_ encoder and decoder for PlantUML_.

Authors
-------
`Dyno Fu`_ and `Rudi Yardley`_.

Resources
---------
- https://plantuml.com/text-encoding
- https://github.com/dougn/python-plantuml/blob/master/plantuml.py#L64
- https://gist.github.com/dyno/94ef6bb9644a88d6981d6a1a9eb70802
- https://gist.github.com/ryardley/64816f5097003a35f9726aab676920d0


.. _PlantUML: https://plantuml.com/
.. _PlantUML Text Encoding format: https://plantuml.com/text-encoding

.. _Dyno Fu: https://github.com/dyno
.. _Rudi Yardley: https://github.com/ryardley
"""
import base64
import string
import zlib

plantuml_alphabet = string.digits + string.ascii_uppercase + string.ascii_lowercase + "-_"
base64_alphabet = string.ascii_uppercase + string.ascii_lowercase + string.digits + "+/"
b64_to_plantuml = bytes.maketrans(base64_alphabet.encode("utf-8"), plantuml_alphabet.encode("utf-8"))
plantuml_to_b64 = bytes.maketrans(plantuml_alphabet.encode("utf-8"), base64_alphabet.encode("utf-8"))


def plantuml_encode(plantuml_text):
    """zlib compress the plantuml text and encode it for the plantuml server"""
    zlibbed_str = zlib.compress(plantuml_text.encode("utf-8"))
    compressed_string = zlibbed_str[2:-4]
    return base64.b64encode(compressed_string).translate(b64_to_plantuml).decode("utf-8")


def plantuml_decode(plantuml_url):
    """decode plantuml encoded url back to plantuml text"""
    data = base64.b64decode(plantuml_url.translate(plantuml_to_b64).encode("utf-8"))
    dec = zlib.decompressobj()  # without check the crc.
    header = b"x\x9c"
    return dec.decompress(header + data).decode("utf-8")


if __name__ == "__main__":  # pragma: nocover
    original = "SyfFKj2rKt3CoKnELR1Io4ZDoSa700=="
    decoded = plantuml_decode(original)
    reencoded = plantuml_encode(decoded)
    print(f"original:  {original}")
    print(f"decoded:   {decoded}")
    print(f"reencoded: {reencoded}")



import logging
import sys


def setup_logging(level=logging.INFO):
    log_format = "%(asctime)-15s [%(name)-15s] %(levelname)-7s: %(message)s"
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)


class ReversibleDict(dict):
    """
    Reverse dictionary lookup.

    http://stupidpythonideas.blogspot.com/2014/07/reverse-dictionary-lookup-and-more-on.html
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rev = {v: k for k, v in self.items()}

    def __delitem__(self, k):
        del self.rev[self[k]]
        super(ReversibleDict, self).__delitem__(k)

    def __setitem__(self, k, v):
        try:
            del self.rev[self[k]]
        except KeyError:
            pass
        super(ReversibleDict, self).__setitem__(k, v)
        self.rev[v] = k

    def lookup(self, v):
        return self.rev[v]



import click

from wireviz_web import create_app
from wireviz_web.server import wireviz_blueprint
from wireviz_web.util import setup_logging


@click.command(
    help="""A wrapper around WireViz for bringing it to the web. 
    Easily document cables and wiring harnesses."""
)
@click.option(
    "--listen",
    type=str,
    required=False,
    default="localhost:3005",
    help="The address the server should listen on." "Defaults to: localhost:3005",
)
@click.option(
    "--url",
    type=str,
    required=False,
    default="",
    help="On wich URL prefix to respond." "Defaults to: /",
)
@click.option(
    "--debug",
    is_flag=True,
    required=False,
    default=False,
    help="Whether to run the server in debug mode",
)
def run(listen: str, url: str, debug: bool):

    # Setup logging.
    setup_logging()

    # Create Flask application.
    app = create_app()

    # Register WireViz-Web component.
    app.register_blueprint(wireviz_blueprint, url_prefix=url)

    # Invoke Flask application.
    host, port = listen.split(":")
    app.run(host=host, port=int(port), debug=debug)



# A wrapper around WireViz for bringing it to the web. Easily document cables and wiring harnesses.
#
# Copyright (C) 2020  Jürgen Key <jkey@arcor.de>
# Copyright (C) 2021  Andreas Motl <andreas.motl@panodata.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import io
import json
import logging
from pathlib import Path
from tempfile import NamedTemporaryFile

import yaml
from flask import Response, send_file
from werkzeug.exceptions import BadRequest, NotAcceptable
from wireviz import wireviz
from wireviz.Harness import Harness
from wireviz.wv_bom import bom_list
from wireviz.wv_helper import tuplelist2tsv

from wireviz_web.plantuml import plantuml_decode
from wireviz_web.util import ReversibleDict

mimetype_type_map: ReversibleDict = ReversibleDict(
    {
        "image/png": "png",
        "image/svg+xml": "svg",
        "text/html": "html",
        "text/plain": "bom.txt",
        "application/json": "bom.json",
    }
)

logger = logging.getLogger(__name__)


def mimetype_to_type(mime_type: str) -> str:
    """
    Translate MIME type to output type.
    For unknown types, raise HTTP Not Acceptable.

    :param mime_type: The MIME type string.
    :return:          The image type string.
    """
    try:
        return mimetype_type_map[mime_type]
    except KeyError:
        raise NotAcceptable(description=f"Output type not acceptable: {mime_type}")


def type_to_mimetype(output_type: str) -> str:
    """
    Translate output type to MIME type.
    For unknown types, raise HTTP Not Acceptable.

    :param output_type: The output type string.
    :return:            The MIME type string.
    """
    try:
        return mimetype_type_map.lookup(output_type)
    except KeyError:
        raise NotAcceptable(description=f"Output type not acceptable: {output_type}")


def decode_plantuml(input_plantuml: str) -> str:
    """
    Decode PlantUML Text Encoding format.

    See also: https://plantuml.com/text-encoding

    :param input_plantuml: The request data, encoded with PlantUML Text Encoding format.
    :return:               The decoded data.
    """
    try:
        return plantuml_decode(input_plantuml)
    except Exception as ex:
        raise BadRequest(description=f"Unable to decode PlantUML Text Encoding format: {ex}")


def wireviz_render(input_yaml: str, output_mimetype: str, output_filename: str) -> Response:
    """
    Render WireViz output and create a Flask Response.

    :param input_yaml:      Input data in WireViz YAML format.
    :param output_mimetype: The designated output format mimetype. Currently available:
                            - image/svg+xml
                            - image/png
                            - text/html
                            - text/plain
                            - application/json
    :param output_filename: The designated output filename.
    :return:                A Flask Response object.
    """

    # Sanity checks.
    if not input_yaml.strip():
        raise BadRequest(description="No input data")

    # Compute output type from MIME type.
    return_type = mimetype_to_type(output_mimetype)

    # Parse WireViz YAML.
    try:
        harness: Harness = wireviz.parse(yaml_input=input_yaml, return_types="harness")
    except Exception as ex:
        message = f"Unable to parse WireViz YAML format: {ex}"
        logger.exception(message)
        raise BadRequest(description=message)

    # Dispatch rendering by designated output type.
    if return_type == "png":
        payload = harness.png

    elif return_type == "svg":
        payload = harness.svg

    elif return_type == "html":
        try:
            tmpfile = NamedTemporaryFile(delete=False)

            # Build list of implicitly created temporary files.
            tempfiles = []
            for suffix in [".gv", ".png", ".svg", ".bom.tsv", ".html"]:
                tempfiles.append(f"{tmpfile.name}{suffix}")

            # Render HTML output.
            harness.output(filename=tmpfile.name, fmt=("png", "svg"))
            with open(f"{tmpfile.name}.html", "rb") as f:
                payload = f.read()

        except Exception as ex:  # pragma: no cover
            message = f"Unable to produce WireViz output: {ex}"
            logger.exception(message)
            raise BadRequest(description=message)

        finally:
            # Clean up temporary files.
            for tempfile in tempfiles:
                try:
                    Path(tempfile).unlink(missing_ok=True)
                except (FileNotFoundError, TypeError):  # pragma: no cover
                    pass

    elif return_type == "bom.txt":
        harness.create_graph()
        bomlist = bom_list(harness.bom())
        payload = tuplelist2tsv(bomlist).encode("utf-8")

    elif return_type == "bom.json":
        harness.create_graph()
        bomlist = bom_list(harness.bom())
        payload = json.dumps(bomlist, indent=2).encode("utf-8")

    # Respond with rendered image.
    return send_file(
        io.BytesIO(payload),
        mimetype=output_mimetype,
        as_attachment=True,
        download_name=output_filename,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wireviz_web']

package_data = \
{'': ['*']}

install_requires = \
['Flask-Cors>=3.0.10,<4.0.0',
 'click>=8,<9',
 'flask-restx>=0.5,<0.6',
 'flask<2.2',
 'werkzeug<2.2',
 'wireviz>=0.3,<0.4']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=3.3.0,<4.0.0']}

entry_points = \
{'console_scripts': ['wireviz-web = wireviz_web.cli:run']}

setup_kwargs = {
    'name': 'wireviz-web',
    'version': '0.4.0',
    'description': 'A wrapper around WireViz for bringing it to the web. Easily document cables and wiring harnesses.',
    'long_description': '###########\nWireViz-Web\n###########\n\n.. image:: https://github.com/daq-tools/wireviz-web/workflows/Tests/badge.svg\n    :target: https://github.com/daq-tools/wireviz-web/actions?workflow=Tests\n.. image:: https://codecov.io/gh/daq-tools/wireviz-web/branch/main/graph/badge.svg\n    :target: https://codecov.io/gh/daq-tools/wireviz-web\n\n.. image:: https://img.shields.io/pypi/v/wireviz-web.svg\n    :target: https://pypi.org/project/wireviz-web/\n.. image:: https://pepy.tech/badge/wireviz-web/month\n    :target: https://pepy.tech/project/wireviz-web\n\n.. image:: https://img.shields.io/pypi/pyversions/wireviz-web.svg\n    :target: https://pypi.org/project/wireviz-web/\n.. image:: https://img.shields.io/pypi/status/wireviz-web.svg\n    :target: https://pypi.org/project/wireviz-web/\n.. image:: https://img.shields.io/github/license/daq-tools/wireviz-web\n    :target: https://github.com/daq-tools/wireviz-web/blob/main/LICENSE\n\n\n*****\nAbout\n*****\nWireViz-Web is a wrapper around the excellent WireViz_ by `Daniel Rojas`_\nfor bringing it to the web.\n\nOriginally, it has been conceived within a `WireViz fork`_ by `Jürgen Key`_.\nFor compatibility with PlantUML_, it includes a `PlantUML Text Encoding format`_\ndecoder by `Dyno Fu`_ and `Rudi Yardley`_.\n\nThanks!\n\n\n*******\nDetails\n*******\n\nWireViz\n=======\n\nWireViz is a tool for easily documenting cables, wiring harnesses and connector pinouts.\nIt takes plain text, YAML-formatted files as input and produces beautiful graphical output\n(SVG, PNG, ...) thanks to Graphviz_.\nIt handles automatic BOM (Bill of Materials) creation and has a lot of extra features.\n\nWireViz-Web\n===========\n\nWireViz-Web wraps WireViz with a REST API using Flask. It also provides specific rendering\nendpoints for PlantUML.\n\n\n*****\nSetup\n*****\n\nInstall prerequisites::\n\n    {apt,brew,dnf,yum,zypper} install python3 graphviz\n\nInstall package::\n\n    pip install wireviz-web\n\n\n*****\nUsage\n*****\n\nRun server::\n\n    wireviz-web\n\nInvoke requests::\n\n    # Acquire WireViz YAML file.\n    wget https://raw.githubusercontent.com/daq-tools/wireviz-web/main/tests/demo01.yaml\n\n    # Render images.\n    http --form http://localhost:3005/render yml_file@demo01.yaml Accept:image/svg+xml\n    http --form http://localhost:3005/render yml_file@demo01.yaml Accept:image/png\n\n    # Render HTML page with SVG image and BOM table.\n    http --form http://localhost:3005/render yml_file@demo01.yaml Accept:text/html\n\n    # Render BOM in TSV format.\n    http --form http://localhost:3005/render yml_file@demo01.yaml Accept:text/plain\n\n    # Render BOM in JSON format.\n    http --form http://localhost:3005/render yml_file@demo01.yaml Accept:application/json\n\n    # Render a PlantUML request.\n    http http://localhost:3005/plantuml/svg/SyfFKj2rKt3CoKnELR1Io4ZDoSa700==\n    http http://localhost:3005/plantuml/png/SyfFKj2rKt3CoKnELR1Io4ZDoSa700==\n\n.. note::\n\n    The ``http`` command used in the examples is the excellent HTTPie_ program.\n\nFor visiting the Swagger OpenAPI spec, go to http://localhost:3005/doc.\n\n\n\n*******************\nProject information\n*******************\n\nContributions\n=============\n\nEvery kind of contribution, feedback, or patch, is much welcome. `Create an\nissue`_ or submit a patch if you think we should include a new feature, or to\nreport or fix a bug.\n\nIn order to follow the general development discussion, please see `Bringing\nWireViz to the Web`_.\n\nDevelopment\n===========\n\nIn order to setup a development environment on your workstation, please head\nover to the `development sandbox`_ documentation. When you see the software\ntests succeed, you should be ready to start hacking.\n\nResources\n=========\n\n- `Source code repository <https://github.com/daq-tools/wireviz-web>`_\n- `Documentation <https://github.com/daq-tools/wireviz-web/blob/main/README.rst>`_\n- `Python Package Index (PyPI) <https://pypi.org/project/wireviz-web/>`_\n\nLicense\n=======\n\nThe project is licensed under the terms of the GNU AGPL license.\n\n\n.. _Bringing WireViz to the Web: https://community.hiveeyes.org/t/bringing-wireviz-to-the-web/3700\n.. _create an issue: https://github.com/daq-tools/wireviz-web/issues\n.. _Daniel Rojas: https://github.com/formatc1702\n.. _development sandbox: https://github.com/daq-tools/wireviz-web/blob/main/doc/sandbox.rst\n.. _Dyno Fu: https://github.com/dyno\n.. _Graphviz: https://www.graphviz.org/\n.. _HTTPie: https://httpie.io/\n.. _Jürgen Key: https://github.com/elbosso\n.. _PlantUML: https://plantuml.com/\n.. _PlantUML Text Encoding format: https://plantuml.com/text-encoding\n.. _Poetry: https://pypi.org/project/poetry/\n.. _Rudi Yardley: https://github.com/ryardley\n.. _WireViz: https://github.com/formatc1702/WireViz\n.. _WireViz fork: https://github.com/elbosso/WireViz\n',
    'author': 'Jürgen Key',
    'author_email': 'jkey@arcor.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://community.hiveeyes.org/t/bringing-wireviz-to-the-web/3700',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# A wrapper around WireViz for bringing it to the web. Easily document cables and wiring harnesses.
#
# Copyright (C) 2020  Jürgen Key <jkey@arcor.de>
# Copyright (C) 2021  Andreas Motl <andreas.motl@panodata.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from pathlib import PurePath

import werkzeug
from flask import Blueprint, Response, request
from flask_restx import Api, Resource, reqparse

from wireviz_web import __version__
from wireviz_web.core import decode_plantuml, mimetype_to_type, type_to_mimetype, wireviz_render

file_upload = reqparse.RequestParser()
file_upload.add_argument(
    "yml_file",
    type=werkzeug.datastructures.FileStorage,
    location="files",
    required=True,
    help="YAML file",
)

wireviz_blueprint = Blueprint("wireviz-web", __name__)
api = Api(
    app=wireviz_blueprint,
    version=__version__,
    title="WireViz-Web",
    description="A wrapper around WireViz for bringing it to the web. " "Easily document cables and wiring harnesses.",
    doc="/doc",
    catch_all_404s=True,
)

ns = api.namespace("", description="WireViz-Web REST API")


@ns.route("/render")
class RenderRegular(Resource):
    @api.expect(file_upload)
    @ns.produces(["image/png", "image/svg+xml"])
    def post(self) -> Response:
        """
        Receive a "multipart/form-data" file upload request with filename "yml_file".
        The "Accept" request header will determine the response image type.

        Examples
        ========
        ::

            # Acquire WireViz YAML file.
            wget https://raw.githubusercontent.com/daq-tools/wireviz-web/main/tests/demo01.yaml

            # Render images.
            http --form http://localhost:3005/render yml_file@demo01.yaml Accept:image/svg+xml
            http --form http://localhost:3005/render yml_file@demo01.yaml Accept:image/png

            # Render HTML page with SVG image and BOM table.
            http --form http://localhost:3005/render yml_file@demo01.yaml Accept:text/html

            # Render BOM in TSV format.
            http --form http://localhost:3005/render yml_file@demo01.yaml Accept:text/plain

            # Render BOM in JSON format.
            http --form http://localhost:3005/render yml_file@demo01.yaml Accept:application/json

        :return: A Flask Response object with the rendered image.
        """

        # The designated output image mime type.
        mimetype = request.headers.get("accept")

        # Read input YAML.
        args = file_upload.parse_args()
        yaml_input = args["yml_file"].read()

        # Determine input- and output file names.
        input_filename = args["yml_file"].filename
        output_filename = PurePath(PurePath(input_filename).stem).with_suffix("." + mimetype_to_type(mimetype)).name

        # Respond with rendered image.
        return wireviz_render(
            input_yaml=yaml_input,
            output_mimetype=mimetype,
            output_filename=output_filename,
        )


@ns.route("/plantuml/<imagetype>/<encoded>")
@ns.param("encoded", "PlantUML Text Encoding format")
class RenderPlantUML(Resource):
    @ns.produces(["image/png", "image/svg+xml"])
    def get(self, imagetype: str, encoded: str) -> Response:
        """
        Receive PlantUML Text Encoding format within URL path.
        The URL prefix will determine the response image type.

        Examples
        ========
        ::

            http http://localhost:3005/png/SyfFKj2rKt3CoKnELR1Io4ZDoSa700==
            http http://localhost:3005/svg/SyfFKj2rKt3CoKnELR1Io4ZDoSa700==

        :return: A Flask Response object with the rendered image.
        """
        mimetype = type_to_mimetype(imagetype)
        yaml_input = decode_plantuml(input_plantuml=encoded)
        return wireviz_render(
            input_yaml=yaml_input,
            output_mimetype=mimetype,
            output_filename=f"rendered.{imagetype}",
        )
